﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace Microsoft.GGTool.Common
{
    public static class GGConvert
    {
        public static Stream ToJson(XmlDocument xData)
        {
            MemoryStream xOut = new MemoryStream();
            int currentDepth = 0;
            string xmlns = "";
            XmlNodeType currentType = XmlNodeType.None;
            bool textPropertyEnabled = false, canBeClosed = true;

            using (var sW = new StreamWriter(xOut, System.Text.Encoding.UTF8, 10000, true))
            using (var writer = new JsonTextWriter(sW))
            using (XmlReader xR = xData.DocumentElement.CreateNavigator().ReadSubtree())
            {
                writer.WriteStartObject();
                while (xR.Read())
                {
                    SKIPREAD:
                    switch (xR.NodeType)
                    {
                        case XmlNodeType.Element:
                            if (!textPropertyEnabled)
                            {
                                // going deeper with no inner text ?
                                if (currentDepth < xR.Depth)
                                    writer.WriteStartObject();
                            }

                            // going up after empty element ?
                            if (currentDepth > xR.Depth)
                            {
                                writer.WriteEndObject();
                                currentDepth--;
                            }

                            if (xmlns != xR.NamespaceURI)
                            {
                                xmlns = xR.NamespaceURI;
                                writer.WritePropertyName("$schema");
                                writer.WriteValue(xR.NamespaceURI);
                            }

                            // preset
                            currentType = xR.NodeType;

                            // attributes check
                            if (HasRelevantAttributes(xR))
                            {
                                writer.WritePropertyName(xR.LocalName);
                                writer.WriteStartObject();
                                textPropertyEnabled = true;
                                canBeClosed = true;
                                currentDepth++;

                                xR.MoveToFirstAttribute();
                                for (int ai = 0; ai < xR.AttributeCount; ai++)
                                {
                                    if (xR.LocalName == "xmlns")
                                    {
                                        xR.MoveToNextAttribute();
                                        continue;
                                    }

                                    writer.WritePropertyName("@" + xR.LocalName);
                                    writer.WriteValue(xR.ReadContentAsString());
                                    xR.MoveToNextAttribute();
                                }
                                xR.MoveToElement();
                                if (xR.IsEmptyElement)
                                {
                                    writer.WriteEndObject();
                                    currentDepth--;
                                }
                            }
                            else
                            {
                                textPropertyEnabled = false;
                                canBeClosed = false;
                                writer.WritePropertyName(xR.LocalName);
                                if (xR.IsEmptyElement)
                                {
                                    writer.WriteStartObject();
                                    writer.WriteEndObject();
                                    currentType = XmlNodeType.EndElement; // pretend there was an endelement, because of the empty element
                                }
                                currentDepth = xR.Depth;
                            }
                            break;
                        case XmlNodeType.EndElement:
                            if (canBeClosed || (currentType == XmlNodeType.EndElement))
                                writer.WriteEndObject();
                            else if (currentType == XmlNodeType.Element)
                                writer.WriteValue(""); // <e></e> type of input
                            currentDepth = xR.Depth;

                            textPropertyEnabled = false;
                            currentType = xR.NodeType;
                            break;
                        case XmlNodeType.Text:
                            currentType = xR.NodeType;
                            if (textPropertyEnabled)
                            {
                                writer.WritePropertyName("#text");
                                if (xR.HasValue)
                                    writer.WriteValue(xR.ReadContentAsString());
                                else
                                    writer.WriteValue("");
                                textPropertyEnabled = false;
                            }
                            else
                            {
                                writer.WriteValue(xR.ReadContentAsString());
                                currentDepth = xR.Depth;
                            }

                            if (xR.NodeType != XmlNodeType.Text)
                                goto SKIPREAD;
                            break;
                        default:
                            currentDepth = xR.Depth;
                            currentType = xR.NodeType;
                            break;
                    }
                }
                writer.WriteEndObject();
            }

            xOut.Seek(0, SeekOrigin.Begin);

            return xOut;
        }

        private static bool HasRelevantAttributes(XmlReader xR)
        {
            bool result = false, keepGoing;

            if (!xR.HasAttributes)
                return false;

            keepGoing = xR.MoveToFirstAttribute();
            while (keepGoing)
            {
                if (xR.LocalName == "xmlns")
                {
                    keepGoing = xR.MoveToNextAttribute();
                    result = false;
                    continue;
                }

                result = true;
                keepGoing = xR.MoveToNextAttribute();
            }
            xR.MoveToElement();

            return result;
        }

        public static XmlDocument ToXml(string jsonData)
        {
            XmlDocument xOut = new XmlDocument();
            string currNameSpace = "";
            int currentDepth = 0;
            using (var ms = new MemoryStream(Encoding.UTF8.GetBytes(jsonData)))            
            using (var sR = new StreamReader(ms))
            using (var reader = new JsonTextReader(sR))
            {
                XmlNode parentNode = null;
                XmlNode lastNode = null;
                int lastLevel = 0;
                Dictionary<string,string> items = new Dictionary<string, string>();
                string propName = "";
                while (reader.Read())
                {

                    switch (reader.TokenType)
                    {
                        case JsonToken.EndObject:
                        case JsonToken.StartObject:

                            if (!string.IsNullOrWhiteSpace(propName))
                            {
                                items.Add(propName, null);
                                propName = "";
                            }
                                                        
                            if (items.Any())
                            {
                                var nSpace = "";
                                if (items.ContainsKey("$schema"))
                                {
                                    nSpace = items["$schema"];
                                }
                                foreach (var nodePair in items.Where(x => x.Key != "$schema" && x.Key[0] != '@'))
                                {

                                    XmlNode node = xOut.CreateElement(nodePair.Key, nSpace);
                                    if (string.IsNullOrWhiteSpace(nSpace))
                                    {
                                        node = xOut.CreateElement(nodePair.Key, currNameSpace);
                                    }

                                    if (!string.IsNullOrWhiteSpace(nodePair.Value))
                                    {
                                        node.InnerText = nodePair.Value;
                                    }

                                    if (currentDepth == 1) // root
                                    {
                                        xOut.AppendChild(node);
                                        lastNode = node;
                                        lastLevel = 1;
                                        currNameSpace = nSpace;
                                        continue;
                                    }
                                    if (currentDepth == lastLevel) // add to same level
                                    {
                                        parentNode.AppendChild(node);
                                        lastNode = node;
                                        continue;
                                    }
                                    if (currentDepth == lastLevel + 1) //add parent
                                    {
                                        lastNode.AppendChild(node);
                                        parentNode = lastNode;
                                        lastNode = node;
                                        lastLevel = currentDepth;
                                        currNameSpace = lastNode.NamespaceURI;
                                        continue;
                                    }

                                    if (currentDepth < lastLevel) // return top
                                    {
                                        do
                                        {
                                            lastLevel--;
                                            lastNode = parentNode;
                                            parentNode = lastNode.ParentNode;
                                        } while (currentDepth - 1 < lastLevel);

                                        lastNode.AppendChild(node);
                                        parentNode = lastNode;
                                        lastNode = node;
                                        lastLevel = currentDepth;
                                        currNameSpace = lastNode.NamespaceURI;
                                        continue;
                                    }
                                }
                                items.Clear();
                            }
                            break;
                        case JsonToken.PropertyName:
                            if (reader.Value != null)
                            {
                                if (string.IsNullOrWhiteSpace(propName))
                                {
                                    propName = reader.Value.ToString();
                                    currentDepth = reader.Depth;
                                }
                            }
                            break;
                        default:
                            if (string.IsNullOrWhiteSpace(propName))
                            {
                                if (reader.Value != null)
                                {
                                    Console.WriteLine("Error - Token: {0}, Value: {1}, Deepth: {2}", reader.TokenType, reader.Value, reader.Depth);
                                }
                                else
                                {
                                    Console.WriteLine("Error - Token: {0}", reader.TokenType);
                                }
                            }
                            else
                            {
                                if (reader.Value != null)
                                {
                                    if (propName == "#text") // is it attribute and add to parent
                                    {
                                        lastNode.InnerText = reader.Value.ToString();
                                    }
                                    else if (propName[0] == '@') // is it attribute and add to parent
                                    {
                                        var attr = xOut.CreateAttribute(propName.Substring(1));
                                        attr.Value = reader.Value.ToString();
                                        lastNode.Attributes.Append(attr);
                                    }
                                    else
                                    {
                                        items.Add(propName, reader.Value.ToString());
                                    }
                                }
                                else
                                {
                                    items.Add(propName, null);
                                }
                                propName = "";
                            }
                            break;
                    }

                }
            }
            return xOut;
        }

    }
}
